import abc
import base64
from typing import Dict, TypeVar, Union, Optional

import attr


TDefault = TypeVar('TDefault')


def lowercase_dict(data: Dict[str, str]) -> Dict[str, str]:
    return {key.lower(): value for key, value in data.items()}


@attr.s
class Headers:
    data: Dict[str, str] = attr.ib(convert=lowercase_dict, default=attr.Factory(dict))

    def get(self, key: str, default: TDefault=None) -> Union[str, TDefault]:
        return self.data.get(key.lower(), default)


@attr.s
class Request:
    method: str = attr.ib()
    url: str = attr.ib()
    headers: Dict[str, str] = attr.ib(default=attr.Factory(lambda: {}))
    body: Optional[bytes] = attr.ib(default=None)


@attr.s
class Response:
    status: int = attr.ib()
    headers: Headers = attr.ib()
    body: bytes = attr.ib(repr=False)


class HTTPSession(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def request(self, request: Request) -> Response:
        raise NotImplementedError()

    @abc.abstractmethod
    async def close(self):
        raise NotImplementedError()


@attr.s
class Engine:
    http_session: HTTPSession = attr.ib()
    start_process = attr.ib()
    sleep = attr.ib()


class Auth(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_headers(self) -> Dict[str, str]:
        raise NotImplementedError()


@attr.s
class BasicAuth(Auth):
    username = attr.ib()
    password = attr.ib()

    def get_headers(self):
        raw_token = f'{self.username}:{self.password}'
        token = base64.b64encode(raw_token.encode('ascii')).decode('ascii')
        return {
            'Authorization': f'Basic {token}'
        }



import base64
import json
from io import BytesIO
from json import JSONDecodeError
from pathlib import Path
from zipfile import ZipFile, ZIP_DEFLATED

from structlog import get_logger

from arsenic import errors
from arsenic.engines import Request, Response, HTTPSession


WEB_ELEMENT = 'element-6066-11e4-a52e-4f735466cecf'


log = get_logger()


def unwrap(value):
    if isinstance(value, dict) and ('ELEMENT' in value or WEB_ELEMENT in value):
        wrapped_id = value.get('ELEMENT', None)
        if wrapped_id:
            return value['ELEMENT']
        else:
            return value[WEB_ELEMENT]

    elif isinstance(value, list):
        return list(unwrap(item) for item in value)
    else:
        return value


def wrap_screen(data):
    if isinstance(data, dict) and 'value' in data and isinstance(data['value'], dict) and 'screen' in data['value']:
        data['value']['screen'] = BytesIO(base64.b64decode(data['value']['screen']))


class Connection:
    def __init__(self, session: HTTPSession, prefix: str):
        self.session = session
        self.prefix = prefix

    async def request(self, *, url: str, method: str, data=None, raw=False):
        if data is None:
            data = {}
        if method not in {'POST', 'PUT'}:
            data = None
        request = Request(
            url=self.prefix + url,
            method=method,
            body=json.dumps(data) if data is not None else None
        )
        log.info('request', request=request)
        response: Response = await self.session.request(request)
        try:
            data = json.loads(response.body)
        except JSONDecodeError as exc:
            log.error('json-decode', body=response.body)
            data = {
                'error': '!internal',
                'message': str(exc),
                'stacktrace': ''
            }
        wrap_screen(data)
        log.info('response', request=request, response=response, data=data)
        errors.check_response(response.status, data)
        if raw:
            return data
        if data:
            return unwrap(data.get('value', None))

    async def upload_file(self, path: Path) -> Path:
        return path

    def prefixed(self, prefix: str) -> 'Connection':
        return Connection(self.session, self.prefix + prefix)


class RemoteConnection(Connection):
    async def upload_file(self, path: Path) -> Path:
        log.info('upload-file', path=path)
        fobj = BytesIO()
        with ZipFile(fobj, 'w', ZIP_DEFLATED) as zf:
            zf.write(path, path.name)
        content = base64.b64encode(fobj.getvalue()).decode('utf-8')
        return await self.request(
            url='/file',
            method='POST',
            data={
                'file': content
            }
        )



import abc
from enum import Enum
from typing import List, Dict, Any, Union

import attr

from arsenic.connection import WEB_ELEMENT
from arsenic.session import Element


TOrigin = Union[Element, 'Origins']


class Origins(Enum):
    viewport = 'viewport'
    pointer = 'pointer'


class Button(Enum):
    left = 0
    middle = 1
    right = 2


class Device(metaclass=abc.ABCMeta):
    id: str = abc.abstractproperty()
    type: str = abc.abstractproperty()
    parameters: Dict[str, str] = abc.abstractproperty()

    def __init__(self):
        self.actions: List[Action] = []

    @abc.abstractmethod
    def encode_actions(self) -> List[Dict[str, Any]]:
        raise NotImplementedError()


@attr.s
class Action:
    type = attr.ib()
    duration = attr.ib()
    data = attr.ib()

    def encode(self):
        return {
            'type': self.type,
            'duration': self.duration,
            **self.data
        }


def encode_origin(origin):
    if isinstance(origin, Element):
        return {WEB_ELEMENT: origin.id}
    elif isinstance(origin, Origins):
        return origin.value
    else:
        raise TypeError()


class Mouse(Device):
    id = 'mouse'
    type = 'pointer'
    parameters = {
        'pointerType': 'mouse'
    }

    def move(self, origin: TOrigin, x: int, y: int, duration: int=250):
        self.actions.append(Action(
            type='pointerMove',
            duration=duration,
            data={
                'origin': encode_origin(origin),
                'x': x,
                'y': y
            }
        ))

    def button_down(self, button: Button=Button.left):
        self.actions.append(Action(
            type='pointerDown',
            duration=0,
            data={
                'button': button.value
            }
        ))

    def button_up(self, button: Button=Button.left):
        self.actions.append(Action(
            type='pointerUp',
            duration=0,
            data={
                'button': button.value
            }
        ))

    def encode_actions(self):
        return [action.encode() for action in self.actions]


class Actions:
    def __init__(self):
        self.mouse = Mouse()
        self.devices: List[Device] = [self.mouse]

    def move_to(self, element: Element) -> 'Actions':
        self.mouse.move(element, 0, 0)
        return self

    def move_by(self, x: int, y: int) -> 'Actions':
        self.mouse.move(Origins.pointer, x, y)
        return self

    def mouse_down(self) -> 'Actions':
        self.mouse.button_down()
        return self

    def mouse_up(self) -> 'Actions':
        self.mouse.button_up()
        return self

    def encode(self):
        return {
            'actions': [
                {
                    'id': device.id,
                    'type': device.type,
                    'parameters': device.parameters,
                    'actions': device.encode_actions(),
                } for device in self.devices
            ]
        }



from asyncio import sleep
from typing import Optional

from aiohttp.client import ClientSession

from . import Engine, Request, Response, Headers, HTTPSession, Auth
from .asyncio import start_process


class Session(HTTPSession):
    def __init__(self, client: ClientSession, auth: Optional[Auth]=None):
        self.client = client
        self.auth = auth

    async def request(self, request: Request) -> Response:
        headers = {
            **request.headers
        }
        if self.auth is not None:
            headers.update(self.auth.get_headers())

        request_context = self.client.request(
            method=request.method,
            url=request.url,
            data=request.body,
            headers=headers
        )

        async with request_context as response:
            body = await response.read()
            status = response.status
            headers = Headers(response.headers)

        return Response(
            status=status,
            body=body,
            headers=headers,
        )

    async def close(self):
        await self.client.close()


async def init_session(auth=None):
    return Session(ClientSession(), auth)


Aiohttp = Engine(
    http_session=init_session,
    start_process=start_process,
    sleep=sleep,
)



import os
from subprocess import DEVNULL
from typing import TextIO, Dict, List, Optional

import attr
from tornado.gen import sleep
from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPError
from tornado.ioloop import IOLoop
from tornado.process import Subprocess

from arsenic.engines import (
    Request, Response, Engine, Headers, HTTPSession,
    Auth,
)


def convert(headers):
    return Headers(dict(headers.get_all()))


class Session(HTTPSession):
    def __init__(self, client: AsyncHTTPClient, auth: Optional[Auth]=None):
        self.client = client
        self.auth = auth

    async def request(self, request: Request) -> Response:
        headers = {**request.headers}
        if self.auth is not None:
            headers.update(**self.auth.get_headers())
        try:
            response = await self.client.fetch(HTTPRequest(
                method=request.method,
                url=request.url,
                body=request.body,
                headers=headers,
            ))
        except HTTPError as exc:
            return Response(
                status=exc.code,
                body=exc.response.body if exc.response else exc.message,
                headers=convert(exc.response.headers) if exc.response else Headers(),
            )
        return Response(
            status=response.code,
            body=response.body,
            headers=convert(response.headers),
        )

    async def close(self):
        self.client.close()


@attr.s
class ProcessContext:
    process = attr.ib()

    async def close(self):
        self.process.proc.terminate()
        io_loop = IOLoop.current()
        def kill():
            try:
                self.process.proc.kill()
            except ProcessLookupError:
                pass
        handle = io_loop.call_later(1, kill)
        await self.process.wait_for_exit(False)
        io_loop.remove_timeout(handle)
        self.process = None
        Subprocess.uninitialize()


async def init_session(auth=None):
    return Session(AsyncHTTPClient(force_instance=True), auth)


async def start_process(cmd: List[str], env: Dict[str, str], log: TextIO) -> ProcessContext:
    if log is os.devnull:
        log = DEVNULL
    return ProcessContext(Subprocess(cmd, stdout=log, stderr=log, env=env))


Tornado = Engine(
    http_session=init_session,
    start_process=start_process,
    sleep=sleep
)



import socket


def free_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind(('0.0.0.0', 0))
        sock.listen(5)
        return sock.getsockname()[1]



from functools import partial
from pathlib import Path
from typing import Awaitable, Callable, Any, List, TYPE_CHECKING

from arsenic.connection import Connection, WEB_ELEMENT
from arsenic.errors import NoSuchElement, OperationNotSupported

if TYPE_CHECKING:
    from arsenic.actions import Actions

UNSET = object()


def escape_value(value: str) -> str:
    if '"' in value and "'" in value:
        parts = value.split('"')
        result = ['concat(']
        for part in parts:
            result.append(f'"{part}"')
            result.append(', \'"\', ')
        result = result[0:-1]
        if value.endswith('"'):
            return ''.join(result) + ')'
        else:
            return ''.join(result[:-1]) + ')'
    elif '"' in value:
        return f"'{value}'"
    else:
        return f'"{value}"'


class Element:
    def __init__(self, id: str, connection: Connection, session: 'Session'):
        self.id = id
        self.connection = connection
        self.session = session

    async def get_text(self) -> str:
        return await self.connection.request(
            url='/text',
            method='GET'
        )

    async def send_keys(self, keys: str):
        await self.connection.request(
            url='/value',
            method='POST',
            data={
                'value': list(keys),
                'text': keys,
            }
        )

    async def send_file(self, path: Path):
        path = await self.session.connection.upload_file(path)
        await self.send_keys(str(path))

    async def clear(self):
        await self.connection.request(
            url='/clear',
            method='POST'
        )

    async def click(self):
        await self.connection.request(
            url='/click',
            method='POST'
        )

    async def is_displayed(self) -> bool:
        return await self.connection.request(
            url='/displayed',
            method='GET'
        )

    async def is_enabled(self) -> bool:
        return await self.connection.request(
            url='/enabled',
            method='GET'
        )

    async def get_attribute(self, name: str) -> str:
        return await self.connection.request(
            url=f'/attribute/{name}',
            method='GET'
        )

    async def select_by_value(self, value: str):
        value = escape_value(value)
        option = await self.get_element(f'option[value={value}]')
        await option.click()

    async def get_element(self, selector: str) -> 'Element':
        element_id = await self.connection.request(
            url='/element',
            method='POST',
            data={
                'using': 'css selector',
                'value': selector,
            }
        )
        return self.session.create_element(element_id)

    async def get_elements(self, selector: str) -> List['Element']:
        element_ids = await self.connection.request(
            url='/elements',
            method='POST',
            data={
                'using': 'css selector',
                'value': selector,
            }
        )
        return [self.session.create_element(element_id) for element_id in element_ids]


TCallback = Callable[..., Awaitable[Any]]
TWaiter = Callable[[int, TCallback], Awaitable[Any]]


class Session:
    element_class = Element

    def __init__(self, connection: Connection, wait=TWaiter, bind: str=''):
        self.connection = connection
        self.bind = bind
        self.wait = wait

    async def get(self, url: str):
        await self.connection.request(
            url='/url',
            method='POST',
            data={
                'url': self.bind + url
            }
        )

    async def get_url(self):
        return await self.connection.request(
            url='/url',
            method='GET'
        )

    async def get_page_source(self):
        return await self.connection.request(
            url='/source',
            method='GET'
        )

    async def get_element(self, selector: str) -> Element:
        element_id = await self.connection.request(
            url='/element',
            method='POST',
            data={
                'using': 'css selector',
                'value': selector
            }
        )
        return self.create_element(element_id)

    async def get_elements(self, selector: str) -> List[Element]:
        result = await self.connection.request(
            url='/elements',
            method='POST',
            data={
                'using': 'css selector',
                'value': selector
            }
        )
        return [self.create_element(element_id) for element_id in result]

    async def wait_for_element(self, timeout: int, selector: str) -> Element:
        return await self.wait(
            timeout,
            partial(self.get_element, selector),
            NoSuchElement
        )

    async def wait_for_element_gone(self, timeout: int, selector: str):
        async def callback():
            try:
                await self.get_element(selector)
            except NoSuchElement:
                return True
            else:
                return False
        return await self.wait(timeout, callback)

    async def add_cookie(self, name, value, *, path=UNSET, domain=UNSET, secure=UNSET, expiry=UNSET):
        cookie = {
            'name': name,
            'value': value
        }
        if path is not UNSET:
            cookie['path'] = path
        if domain is not UNSET:
            cookie['domain'] = domain
        if secure is not UNSET:
            cookie['secure'] = secure
        if expiry is not UNSET:
            cookie['expiry'] = expiry
        await self.connection.request(
            url='/cookie',
            method='POST',
            data={
                'cookie': cookie
            }
        )

    async def get_cookie(self, name):
        return await self.connection.request(
            url=f'/cookie/{name}',
            method='GET'
        )

    async def get_all_cookies(self):
        return await self.connection.request(
            url='/cookie',
            method='GET'
        )

    async def delete_cookie(self, name):
        await self.connection.request(
            url=f'/cookie/{name}',
            method='DELETE'
        )

    async def delete_all_cookies(self):
        await self.connection.request(
            url='/cookie',
            method='DELETE'
        )

    async def execute_script(self, script, *args):
        return await self.connection.request(
            url='/execute/sync',
            method='POST',
            data={
                'script': script,
                'args': list(args)
            }
        )

    async def set_window_size(self, width: int, height: int, handle: str='current'):
        return await self.connection.request(
            url='/window/rect',
            method='POST',
            data={
                'width': width,
                'height': height,
                'windowHandle': handle,
            }
        )

    async def get_window_size(self, handle: str='current'):
        return await self.connection.request(
            url='/window/rect',
            method='GET',
            data={
                'windowHandle': handle,
            }
        )

    async def get_alert_text(self) -> str:
        return await self.connection.request(
            url='/alert/text',
            method='GET'
        )

    async def send_alert_text(self, value: str):
        return await self.connection.request(
            url='/alert/text',
            method='POST',
            data={
                'text': value
            }
        )

    async def dismiss_alert(self):
        return await self.connection.request(
            url='/alert/dismiss',
            method='POST'
        )

    async def accept_alert(self):
        return await self.connection.request(
            url='/alert/accept',
            method='POST'
        )

    async def perform_actions(self, actions: 'Actions'):
        return await self.connection.request(
            url='/actions',
            method='POST',
            data=actions.encode()
        )

    async def close(self):
        await self.connection.request(
            url='',
            method='DELETE'
        )

    def create_element(self, element_id):
        return self.element_class(
            element_id,
            self.connection.prefixed(f'/element/{element_id}'),
            self
        )


class CompatSession(Session):
    async def set_window_size(self, width: int, height: int,
                              handle: str = 'current'):
        return await self.connection.request(
            url=f'/window/{handle}/size',
            method='POST',
            data={
                'width': width,
                'height': height
            }
        )

    async def get_window_size(self, handle: str = 'current'):
        return await self.connection.request(
            url=f'/window/{handle}/size',
            method='GET'
        )

    async def execute_script(self, script, *args):
        return await self.connection.request(
            url='/execute',
            method='POST',
            data={
                'script': script,
                'args': list(args)
            }
        )

    async def _pointer_down(self, device, action):
        await self.connection.request(
            url='/buttondown',
            method='POST',
            data=action.data
        )

    async def _pointer_up(self, device, action):
        await self.connection.request(
            url='/buttonup',
            method='POST',
            data=action.data
        )

    async def _pointer_move(self, device, action):
        origin = action.data['origin']
        if origin == 'pointer':
            await self.connection.request(
                url='/moveto',
                method='POST',
                data={
                    'xoffset': action.data['x'],
                    'yoffset': action.data['y'],
                }
            )
        elif WEB_ELEMENT in origin:
            await self.connection.request(
                url='/moveto',
                method='POST',
                data={
                    'element': origin[WEB_ELEMENT],
                }
            )
        else:
            raise OperationNotSupported(f'Cannot move using origin {origin}')

    action_executors = {
        ('pointer', 'pointerDown'): _pointer_down,
        ('pointer', 'pointerUp'): _pointer_up,
        ('pointer', 'pointerMove'): _pointer_move
    }

    async def perform_actions(self, actions: 'Actions'):
        for device in actions.devices:
            for action in device.actions:
                key = (device.type, action.type)
                try:
                    executor = self.action_executors[key]
                except KeyError:
                    raise OperationNotSupported(
                        f'Action {action.type} of device type {device.type} is '
                        f'not supported by this session'
                    )
                await executor(self, device, action)



from structlog import get_logger

log = get_logger()


class ArsenicError(Exception):
    pass


class OperationNotSupported(ArsenicError):
    pass


class WebdriverError(ArsenicError):
    def __init__(self, message, screen, stacktrace):
        self.message = message
        self.screen = screen
        self.stacktrace = stacktrace
        super().__init__(message)


class UnknownArsenicError(ArsenicError):
    pass


class ArsenicTimeout(ArsenicError):
    pass


CODES = {}


def get(error_code):
    return CODES.get(error_code, UnknownArsenicError)


def create(error_name, *error_codes):
    name = ''.join(bit.capitalize() for bit in error_name.split(' '))
    cls = type(name, (WebdriverError,), {})
    CODES[error_name] = cls
    for code in error_codes:
        CODES[code] = cls
    return cls


NoSuchElement = create('no such element', 7)
NoSuchFrame = create('no such frame', 8)
UnknownCommand = create('unknown command', 9)
StaleElementReference = create('stale element reference', 10)
ElementNotVisible = create('element not visible', 11)
InvalidElementState = create('invalid element state', 12)
UnknownError = create('unknown error', 13)
ElementNotInteractable = create('element not interactable')
ElementIsNotSelectable = create('element is not selectable', 15)
JavascriptError = create('javascript error', 17)
Timeout = create('timeout', 21)
NoSuchWindow = create('no such window', 23)
InvalidCookieDomain = create('invalid cookie domain', 24)
UnableToSetCookie = create('unable to set cookie', 25)
UnexpectedAlertOpen = create('unexpected alert open', 26)
NoSuchAlert = create('no such alert', 27)
ScriptTimeout = create('script timeout', 28)
InvalidElementCoordinates = create('invalid element coordinates', 29)
IMENotAvailable = create('ime not available', 30)
IMEEngineActivationFailed = create('ime engine activation failed', 31)
InvalidSelector = create('invalid selector', 32)
MoveTargetOutOfBounds = create('move target out of bounds', 34)


def _value_or_default(obj, key, default):
    return obj[key] if key in obj else default


def check_response(status, data):
    if status >= 400:
        error = None
        if 'status' in data:
            error = data['status']
        elif 'error' in data:
            error = data['error']
        elif 'state' in data:
            error = data['state']

        if 'value' in data:
            data = data['value']
        if error is None and 'error' in data:
            error = data['error']
        message = data.get('message', None)
        stacktrace = data.get('stacktrace', None)
        screen = data.get('screen', None)
        exception_class = get(error)
        log.error('error', type=exception_class, message=message, stacktrace=stacktrace, data=data, status=status)
        raise exception_class(message, screen, stacktrace)



import abc
import os
from typing import List, TextIO

import attr

from arsenic.connection import Connection
from arsenic.engines import Request, Engine, Auth, BasicAuth
from arsenic.utils import free_port
from arsenic.webdriver import WebDriver


async def subprocess_based_service(engine: Engine,
                                   cmd: List[str],
                                   service_url: str,
                                   log_file: TextIO) -> WebDriver:
    closers = []
    try:
        process = await engine.start_process(
            cmd,
            os.environ,
            log_file
        )
        closers.append(process.close)
        session = await engine.http_session()
        closers.append(session.close)
        request = Request(
            url=service_url + '/status',
            method='GET'
        )
        count = 0
        while True:
            try:
                await session.request(request)
                break
            except:
                count += 1
                if count > 30:
                    raise Exception('not starting?')
                await engine.sleep(0.5)
        return WebDriver(
            engine,
            Connection(session, service_url),
            closers,
        )
    except:
        for closer in reversed(closers):
            await closer()
        raise


class Service(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def start(self, engine):
        raise NotImplementedError()

    def run(self, engine):
        return ServiceContext(self, engine)


@attr.s
class ServiceContext:
    service: Service = attr.ib()
    engine = attr.ib()
    driver = attr.ib(default=None)

    async def __aenter__(self):
        self.driver = await self.service.start(self.engine)
        return self.driver

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.driver.close()
        self.driver = None


@attr.s
class Geckodriver(Service):
    log_file = attr.ib(default=os.devnull)
    binary = attr.ib(default='geckodriver')

    async def start(self, engine):
        port = free_port()
        return await subprocess_based_service(
            engine,
            [self.binary, '--port', str(port)],
            f'http://localhost:{port}',
            self.log_file
        )


def auth_or_string(value):
    if isinstance(value, Auth):
        return value
    elif isinstance(value, str) and value.count(':') == 1:
        username, password = value.split(':')
        return BasicAuth(username, password)
    else:
        raise TypeError()


@attr.s
class Remote(Service):
    url = attr.ib()
    auth = attr.ib(default=None, convert=attr.converters.optional(auth_or_string))

    async def start(self, engine):
        closers = []
        try:
            session = await engine.http_session(self.auth)
            closers.append(session.close)
            return WebDriver(engine, Connection(session, self.url), closers)
        except:
            for closer in reversed(closers):
                await closer()
            raise


@attr.s
class PhantomJS(Service):
    log_file = attr.ib(default=os.devnull)
    binary = attr.ib(default='phantomjs')

    async def start(self, engine):
        port = free_port()
        return await subprocess_based_service(
            engine,
            [self.binary, f'--webdriver={port}'],
            f'http://localhost:{port}/wd/hub',
            self.log_file,
        )



import asyncio
import os
from asyncio.subprocess import DEVNULL
from typing import TextIO, Dict, List

import attr


@attr.s
class ProcessContext:
    process = attr.ib()

    async def close(self):
        self.process.terminate()
        try:
            await asyncio.wait_for(self.process.communicate(), 1)
        except asyncio.futures.TimeoutError:
            self.process.kill()
        try:
            await asyncio.wait_for(self.process.communicate(), 1)
        except asyncio.futures.TimeoutError:
            pass


async def start_process(cmd: List[str], env: Dict[str, str], log: TextIO) -> ProcessContext:
    if log is os.devnull:
        log = DEVNULL
    process = await asyncio.create_subprocess_exec(
        *cmd,
        env=env,
        stdout=log,
        stderr=log,
    )
    return ProcessContext(process)
